import { resolve, join } from 'path'
import {
  getCurrentInstance,
  onServerPrefetch as onServerPrefetch$1,
  ref,
  isRef,
  onBeforeMount,
  reactive,
  toRefs,
} from '@vue/composition-api'
export {
  computed,
  createComponent,
  createElement,
  defineComponent,
  getCurrentInstance,
  inject,
  isRef,
  onActivated,
  onBeforeMount,
  onBeforeUnmount,
  onBeforeUpdate,
  onDeactivated,
  onErrorCaptured,
  onMounted,
  onUnmounted,
  onUpdated,
  provide,
  reactive,
  ref,
  set,
  toRefs,
  watch,
  watchEffect,
} from '@vue/composition-api'
import Vue from 'vue'

const ssrRefFunctions = new WeakMap()
const isPending = new WeakMap()
function onServerPrefetch(cb) {
  const vm = getCurrentInstance()
  if (!vm) throw new Error('ssrRef must be called within setup()')
  const pending = isPending.get(vm) || 0
  isPending.set(vm, pending + 1)
  onServerPrefetch$1(async () => {
    await cb()
    const pending = isPending.get(vm) || 0
    if (pending <= 1) {
      const fn = ssrRefFunctions.get(vm) || []
      await Promise.all(fn.map(p => p()))
    } else {
      isPending.set(vm, pending - 1)
    }
  })
}
function onFinalServerPrefetch(cb) {
  if (!process.server) return
  const vm = getCurrentInstance()
  if (!vm)
    throw new Error('onFinalServerPrefetch must be called within setup()')
  const fn = ssrRefFunctions.get(vm)
  if (!fn) {
    ssrRefFunctions.set(vm, [cb])
  } else {
    fn.push(cb)
  }
}

function getValue(value) {
  if (value instanceof Function) return value()
  return value
}
let data = {}
function setSSRContext(ssrContext) {
  data = Object.assign({}, {})
  ssrContext.nuxt.ssrRefs = data
}
function clone(obj) {
  if (typeof obj === 'object') {
    return JSON.parse(JSON.stringify(obj))
  } else {
    return obj
  }
}
/**
 * Creates a Ref that is in sync with the client.
 */
const ssrRef = (value, key) => {
  var _a, _b, _c
  if (!key) {
    throw new Error(
      "You must provide a key. You can have it generated automatically by adding 'nuxt-composition-api/babel' to your Babel plugins."
    )
  }
  if (process.client) {
    return ref(
      (_c =
        (_b =
          (_a = window.__NUXT__) === null || _a === void 0
            ? void 0
            : _a.ssrRefs) === null || _b === void 0
          ? void 0
          : _b[key]) !== null && _c !== void 0
        ? _c
        : getValue(value)
    )
  }
  const val = getValue(value)
  const initVal = clone(val)
  const _ref = ref(val)
  onFinalServerPrefetch(() => {
    if (value instanceof Function || initVal !== _ref.value)
      data[key] = _ref.value
  })
  return _ref
}

const useAsync = (cb, key) => {
  if (!key) {
    throw new Error(
      "You must provide a key. You can have it generated automatically by adding 'nuxt-composition-api/babel' to your Babel plugins."
    )
  }
  const _ref = isRef(key) ? key : ssrRef(null, key)
  if (!_ref.value) {
    const p = cb()
    if (p instanceof Promise) {
      if (process.server) {
        onServerPrefetch(async () => {
          _ref.value = await p
        })
      } else {
        // eslint-disable-next-line
        p.then(res => (_ref.value = res))
      }
    } else {
      _ref.value = p
    }
  }
  return _ref
}

/**
 * @deprecated
 * Recommend using `useContext` instead
 */
const withContext = callback => {
  const vm = getCurrentInstance()
  if (!vm) throw new Error('This must be called within a setup function.')
  callback(vm.$nuxt.context)
}
const useContext = () => {
  const vm = getCurrentInstance()
  if (!vm) throw new Error('This must be called within a setup function.')
  return vm.$nuxt.context
}

function normalizeError(err) {
  let message
  if (!(err.message || typeof err === 'string')) {
    try {
      message = JSON.stringify(err, null, 2)
    } catch (e) {
      message = `[${err.constructor.name}]`
    }
  } else {
    message = err.message || err
  }
  return {
    ...err,
    message,
    statusCode:
      err.statusCode ||
      err.status ||
      (err.response && err.response.status) ||
      500,
  }
}
const fetches = new WeakMap()
const isSsrHydration = vm => {
  var _a, _b, _c
  return (_c =
    (_b = (_a = vm.$vnode) === null || _a === void 0 ? void 0 : _a.elm) ===
      null || _b === void 0
      ? void 0
      : _b.dataset) === null || _c === void 0
    ? void 0
    : _c.fetchKey
}
const nuxtState = process.client && window.__NUXT__
function registerCallback(vm, callback) {
  const callbacks = fetches.get(vm) || []
  fetches.set(vm, [...callbacks, callback])
}
async function callFetches() {
  const fetchesToCall = fetches.get(this)
  if (!fetchesToCall) return
  this.$nuxt.nbFetching++
  this.$fetchState.pending = true
  this.$fetchState.error = null
  this._hydrated = false
  let error = null
  const startTime = Date.now()
  try {
    await Promise.all(fetchesToCall.map(fetch => fetch(this)))
  } catch (err) {
    error = normalizeError(err)
  }
  const delayLeft = (this._fetchDelay || 0) - (Date.now() - startTime)
  if (delayLeft > 0) {
    await new Promise(resolve => setTimeout(resolve, delayLeft))
  }
  this.$fetchState.error = error
  this.$fetchState.pending = false
  this.$fetchState.timestamp = Date.now()
  this.$nextTick(() => this.$nuxt.nbFetching--)
}
async function serverPrefetch(vm) {
  if (!vm._fetchOnServer) {
    return
  }
  // Call and await on $fetch
  vm.$fetchState =
    vm.$fetchState ||
    Vue.observable({
      error: null,
      pending: false,
      timestamp: 0,
    })
  try {
    await callFetches.call(vm)
  } catch (err) {
    vm.$fetchState.error = normalizeError(err)
  }
  vm.$fetchState.pending = false
  // Define an ssrKey for hydration
  vm._fetchKey = vm.$ssrContext.nuxt.fetch.length
  // Add data-fetch-key on parent element of Component
  if (!vm.$vnode.data) vm.$vnode.data = {}
  const attrs = (vm.$vnode.data.attrs = vm.$vnode.data.attrs || {})
  attrs['data-fetch-key'] = vm._fetchKey
  // Add to ssrContext for window.__NUXT__.fetch
  vm.$ssrContext.nuxt.fetch.push(
    vm.$fetchState.error
      ? { _error: vm.$fetchState.error }
      : JSON.parse(JSON.stringify(vm._data))
  )
}
const useFetch = callback => {
  var _a
  const vm = getCurrentInstance()
  if (!vm) throw new Error('This must be called within a setup function.')
  vm.$fetchState =
    vm.$fetchState ||
    Vue.observable({
      error: null,
      pending: false,
      timestamp: 0,
    })
  vm.$fetch = callFetches.bind(vm)
  registerCallback(vm, callback)
  if (typeof vm.$options.fetchOnServer === 'function') {
    vm._fetchOnServer = vm.$options.fetchOnServer.call(vm) !== false
  } else {
    vm._fetchOnServer = vm.$options.fetchOnServer !== false
  }
  onServerPrefetch(() => serverPrefetch(vm))
  onBeforeMount(() => {
    if (!vm._hydrated) {
      return callFetches.call(vm)
    }
  })
  if (process.server || !isSsrHydration(vm)) return
  // Hydrate component
  vm._hydrated = true
  vm._fetchKey = +((_a = vm.$vnode.elm) === null || _a === void 0
    ? void 0
    : _a.dataset.fetchKey)
  const data = nuxtState.fetch[vm._fetchKey]
  // If fetch error
  if (data && data._error) {
    vm.$fetchState.error = data._error
    return
  }
  onBeforeMount(() => {
    // Merge data
    for (const key in data) {
      try {
        Vue.set(vm, key, data[key])
      } catch (e) {
        if (process.env.NODE_ENV === 'development')
          // eslint-disable-next-line
          console.warn(`Could not hydrate ${key}.`)
      }
    }
  })
}

function createEmptyMeta() {
  return {
    title: undefined,
    titleTemplate: undefined,
    htmlAttrs: undefined,
    headAttrs: undefined,
    bodyAttrs: undefined,
    base: undefined,
    meta: [],
    link: [],
    style: [],
    script: [],
    noscript: [],
    changed: undefined,
    afterNavigation: undefined,
  }
}
function useHead(init = {}) {
  const meta = reactive({
    ...createEmptyMeta(),
    ...init,
  })
  return {
    head: () => meta,
    useMeta: () => toRefs(meta),
  }
}

const compositionApiModule = function () {
  const libRoot = resolve(__dirname, '..')
  const { dst } = this.addTemplate({
    src: resolve(libRoot, 'lib', 'plugin.js'),
    fileName: join('composition-api', 'plugin.js'),
    options: {},
  })
  this.options.build = this.options.build || {}
  this.options.build.babel = this.options.build.babel || {}
  this.options.build.babel.plugins = this.options.build.babel.plugins || []
  this.options.build.babel.plugins.push(join(__dirname, 'babel'))
  this.options.build.transpile = this.options.build.transpile || []
  this.options.build.transpile.push(/nuxt-composition-api/)
  this.options.plugins = this.options.plugins || []
  this.options.plugins.push(resolve(this.options.buildDir || '', dst))
}
// eslint-disable-next-line
const meta = require('../package.json')

export default compositionApiModule
export {
  meta,
  onServerPrefetch,
  setSSRContext,
  ssrRef,
  useAsync,
  useContext,
  useFetch,
  useHead,
  withContext,
}
